# coding: utf-8

from konvertek.parsing import get_args
from konvertek.process import video_processing, list_extensions_processing, problem_info_processing


def main():
    args = get_args()
    if args.command == "convert":
        video_processing(args)
    elif args.command == "list_extensions":
        list_extensions_processing(args)
    elif args.command == "problem_info":
        problem_info_processing(args)
    else:
        print("Failed successfully (main). ")


if __name__ == "__main__":
    main()



# coding: utf-8

import os
import argparse
import shutil
from tqdm import tqdm
from ksupk import get_files_list, get_dirs_needed_for_files, mkdir_with_p
from konvertek.progress_handler import ProgressHandler
from konvertek.ffmpeg_handler import get_video_ext, get_audio_ext, get_image_ext, transcode_video, do_ffmpeg_vf_flag
from konvertek.ffmpeg_handler import get_video_bitrate, parse_bitrate


def video_processing(args: argparse.Namespace):
    ph = ProgressHandler(args.progress_file_path)
    video_ext = get_video_ext()
    d = ph.create_files_4_progress(args.folder_path_in, args.folder_path_out)

    needed_dirs = get_dirs_needed_for_files([os.path.join(str(args.folder_path_out), file_i) for file_i in d.keys()])
    for dir_i in needed_dirs:
        mkdir_with_p(dir_i)

    d_with_media = {}
    for k_i in d:
        file_i = k_i
        if str(os.path.splitext(file_i)[1]).lower() in video_ext:
            d_with_media[file_i] = d[file_i]

    ph.add_files(d)

    ok_files, shitty_files = [], []
    pbar = tqdm(d)
    for k_i in pbar:
        file_i = k_i
        pbar.set_postfix({"current file": f"{file_i}"})
        if not ph.file_status(file_i):
            if file_i in d_with_media:
                vf = do_ffmpeg_vf_flag(args.resolution, args.fps, args.interpolation)
                file_i_in, file_i_out = ph.get_file_in_out(file_i)

                bitrate_i = args.bitrate
                if args.bitrate is not None and not args.force_bitrate:
                    cur_bitrate = get_video_bitrate(file_i_in)
                    target_bitrate = parse_bitrate(bitrate_i)
                    # print(f"1. {cur_bitrate}_cur vs {target_bitrate}_target")
                    if cur_bitrate is not None and target_bitrate > cur_bitrate:
                        bitrate_i = str(cur_bitrate)

                tvr = transcode_video(file_i_in, file_i_out,
                                      v_codec=args.v_codec, vf=vf,
                                      bitrate=bitrate_i,
                                      maxbitrate=args.bitrate if args.maxbitrate else None,
                                      overwrite=args.replace,
                                      print_command=args.print_command,)
                if tvr is None:
                    ph.update(file_i, True, None)
                    ok_files.append(file_i)
                    # print(f"2. {get_video_bitrate(file_i_out)}")
                else:
                    ph.update(file_i, False, tvr)
                    shitty_files.append(file_i)
                    if args.stop_if_error:
                        print(f"Error occurred: \n{tvr}")
                        exit(-1)
            else:
                shutil.copy(*ph.get_file_in_out(file_i))
                ph.update(file_i, True, None)
                ok_files.append(file_i)

    print(f"Total files: {len(ok_files)+len(shitty_files)}, processed: {len(ok_files)}, failed: {len(shitty_files)}")
    if len(shitty_files) > 0:
        print(f"This files were not processed (transcode): ")
        for file_i in shitty_files:
            print(f"- \"{file_i}\"")
        print(f"\n For more details type command: \n> konvertek problem_info {args.progress_file_path}")


def problem_info_processing(args: argparse.Namespace):
    if not os.path.isfile(args.progress_file_path):
        print(f"No such file: {args.progress_file_path}")
    ph = ProgressHandler(args.progress_file_path)
    files, errors = ph.get_files(), ph.get_errors()
    no_processed = []
    for file_i in files:
        if not ph.file_status(file_i):
            no_processed.append(file_i)
    if len(no_processed) > 0:
        print(f"These files were not processed/transcoded: ")
        for file_i in no_processed:
            print(f"- \"{file_i}\"")
    if len(errors.keys()) > 0:
        for k_i in errors:
            print(f"\n{'='*80}\nFile \"{k_i}\": \n{errors[k_i]}{'='*80}")
    else:
        print("No problems output from ffmpeg. ")


def list_extensions_processing(args: argparse.Namespace):
    files = get_files_list(args.folder_path)
    res = set()
    for file_i in files:
        # file_i_name = os.path.basename(file_i)
        # n, e = os.path.splitext(file_i_name)
        # n, e = str(n), str(e).lower()
        # res.add(e)
        res.add(str(os.path.splitext(file_i)[1]).lower())

    print(f"All file extensions encountered: {res}")
    buff1, buff_ext1 = [], get_video_ext()
    for ext_i in res:
        if ext_i in buff_ext1:
            buff1.append(ext_i)
    if len(buff1) > 0:
        print(f"Video formats: {set(buff1)}")
    buff2, buff_ext2 = [], get_audio_ext()
    for ext_i in res:
        if ext_i in buff_ext2:
            buff2.append(ext_i)
    if len(buff2) > 0:
        print(f"Audio formats: {set(buff2)}")
    buff3, buff_ext3 = [], get_image_ext()
    for ext_i in res:
        if ext_i in buff_ext3:
            buff3.append(ext_i)
    if len(buff3) > 0:
        print(f"Image formats: {set(buff3)}")
    buff = []
    for ext_i in res:
        if ext_i in buff_ext1 or ext_i in buff_ext2 or ext_i in buff_ext3:
            pass
        else:
            buff.append(ext_i)
    if len(buff) > 0 and len(buff1)+len(buff2)+len(buff3) > 0:
        print(f"Other formats: {set(buff)}")



# coding: utf-8

import ffmpeg
import os


def transcode_video(input_file, output_file,
                    v_codec: str = None, vf: str = None,
                    bitrate: str = None, maxbitrate: str = None,
                    overwrite: bool = True, print_command: bool = False) -> str | None:
    if not overwrite and os.path.isfile(output_file):
        return None
    out_params = {
        "map": 0,
        "c:a": "copy",
        "c:s": "copy",
    }
    if v_codec is not None:
        out_params["c:v"] = v_codec
    if vf is not None:
        out_params["vf"] = vf
    if bitrate is not None:
        out_params["b:v"] = bitrate
    if maxbitrate is not None:
        out_params["maxrate"] = maxbitrate

    if print_command:
        try:
            process = (
                ffmpeg.input(input_file).output(
                    output_file, **out_params
                ).global_args(*[])
            )
            command = process.compile()
            print(' '.join(command))
        except ffmpeg.Error as e:
            print("Cannot print ffmpeg-command. ")

    try:
        ffmpeg.input(input_file).output(
            output_file, **out_params
        ).global_args(*[]).run(overwrite_output=overwrite, capture_stderr=True, capture_stdout=True)

    except ffmpeg.Error as e:
        return e.stderr.decode("utf8")
        # return e.stdout.decode("utf8")

    return None


def remove_p_from_resolution(s: str) -> int:
    return int(s[:len(s) - 1])


def do_ffmpeg_vf_flag(resolution: str | None, fps: int | None, interpolation: str | None) -> str | None:
    s = ""
    res_to_width = {720: 1280, 1080: 1920, 1440: 2560, 2160: 3840, 4320: 7680, 8640: 15360}
    # "vf": "scale='if(gt(iw,ih),1280,-2):if(gt(iw,ih),-2,720)'",
    # "vf": "scale=-2:'if(gt(ih,720),720,ih)'"
    # "scale=-2:'if(gt(ih,720),720,ih)',fps=60:flags=blend"
    # -vf "minterpolate='fps=60:mi_mode=mci:me=bilat'"
    # scale=-2:'if(gt(ih,720),720,ih)',minterpolate=fps=300:mi_mode=accurate
    if resolution is not None:
        res_int = remove_p_from_resolution(resolution)
        if res_int in res_to_width:
            needed_w = res_to_width[res_int]
            s += f"scale='if(gt(iw,{needed_w}),{needed_w},iw)':-2"
        else:
            s += f"scale=-2:'if(gt(ih,{res_int}),{res_int},ih)'"
    if fps is not None:
        if s != "":
            s += ","
        if interpolation is not None:
            s += f"minterpolate=fps={fps}:mi_mode='{interpolation}'"
        else:
            s += f"fps={fps}"
    return s if s != "" else None


def get_video_ext() -> set:
    res = {".mp4", ".m4v", ".mkv", ".mk3d", ".mka", ".webm", ".avi", ".mov", ".wmv", ".wma", ".asf", ".ts", ".m2ts",
           ".flv",
           ".3gp", ".3g2", ".rm", ".rmvb", ".divx", ".mxf", ".gxf", ".nut", ".psp"}

    return res


def get_audio_ext() -> set:
    res = {".mp3", ".wav", ".aac", ".flac", ".ogg", ".oga", ".amr"}
    return res


def get_image_ext() -> set:
    res = {".jpg", ".jpeg", ".png", ".bmp", ".tiff", ".tif", ".gif", ".webp", ".svg", ".heif", ".hei", ".ico",
           ".ppm", ".pgm", ".pbm", ".pnm",
           ".pcx", ".dds", ".tga", ".icb", ".vda", ".vst", ".exr", ".jp2", ".j2k", ".pgf", ".xbm"}
    return res


def get_video_bitrate(file_path: str) -> int or None:
    try:
        probe = ffmpeg.probe(file_path)

        # first video stream
        video_streams = [stream for stream in probe['streams'] if stream['codec_type'] == 'video']

        if not video_streams:
            return None

        video_stream = video_streams[0]

        bitrate = video_stream.get('bit_rate', None)

        if bitrate:
            return int(bitrate)  # bit/sec
        else:
            return None
    except ffmpeg.Error as e:
        return None


def parse_bitrate(bitrate_str: str) -> int:
    """
    Преобразует строку битрейта (например, '2M', '500K') в целое число (бит/сек).
    """

    suffix_multipliers = {
        'K': 10 ** 3,
        'M': 10 ** 6,
        'G': 10 ** 9
    }

    if bitrate_str[-1] in suffix_multipliers:
        number_part = bitrate_str[:-1]
        suffix = bitrate_str[-1].upper()

        try:
            return int(float(number_part) * suffix_multipliers[suffix])
        except ValueError:
            raise ValueError(f"Incorrect bitrate format: \"{bitrate_str}\"")
    else:
        try:
            return int(bitrate_str)
        except ValueError:
            raise ValueError(f"Incorrect bitrate format: \"{bitrate_str}\"")



# coding: utf-8

from ksupk import singleton_decorator, restore_json, save_json
from ksupk import get_files_list
from threading import Lock
import os


@singleton_decorator
class ProgressHandler:
    def __init__(self, json_path: str):
        self.__lock = Lock()
        self.__path = json_path

        self.__old_containers = [".avi"]

        if not os.path.isfile(json_path):
            self.create_new()
        self.db = restore_json(json_path)

    def flush(self):
        with self.__lock:
            save_json(self.__path, self.db)

    def add_files(self, files: dict):
        for k_i in files:
            self.db["files"][k_i] = files[k_i]
        self.flush()

    def update(self, file: str, doned: bool, error_text: str | None):
        self.db["files"][file]["status"] = doned
        if error_text is not None:
            self.db["errors"][file] = error_text
        else:
            self.db["errors"].pop(file, None)
        self.flush()

    def file_status(self, file: str) -> bool:
        return self.db["files"][file]["status"]

    def get_file_in_out(self, file: str) -> tuple:
        return self.db["files"][file]["in"], self.db["files"][file]["out"]

    def get_db(self) -> dict:
        return self.db.copy()

    def set_error(self, file: str, error_text: str):
        self.db["errors"][file] = error_text
        self.flush()

    def get_errors(self) -> dict:
        return self.db["errors"]

    def get_files(self) -> dict:
        return self.db["files"]

    def create_new(self):
        d = {
            "files": {},
            "errors": {}
        }
        self.db = d
        self.flush()

    def create_files_4_progress(self, dir_in: str, dir_out: str) -> dict:
        d = {}
        files = get_files_list(os.path.abspath(dir_in))
        files = [os.path.relpath(file_i, dir_in) for file_i in files]
        for file_i in files:
            status = False
            if file_i in self.db["files"]:
                if "status" in self.db["files"][file_i]:
                    status = self.db["files"][file_i]["status"]
            out_file_path = self.define_save_path(dir_out, file_i)
            d[file_i] = {"in": os.path.join(dir_in, file_i),
                         "out": out_file_path,
                         "status": status
                         }
        return d

    def define_save_path(self, out_folder: str, file: str) -> str:
        if str(os.path.splitext(file)[1]).lower() in self.__old_containers:
            p, e = os.path.splitext(file)
            return os.path.join(out_folder, f"{p}.mp4")
        else:
            return os.path.join(out_folder, file)



# coding: utf-8

import argparse
import os
from konvertek.__init__ import __version__


def get_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="konvertek. Convert media from folder.")

    parser.add_argument("--version", action="version", version=f"V{__version__}", help="Check version. ")

    subparsers = parser.add_subparsers(dest='command', required=True)

    # convert
    main_parser = subparsers.add_parser('convert', help='transcode')
    main_parser.add_argument('folder_path_in', type=is_dir_path,
                             help="Path to source directory. konvertek will not change inside anything. "
                                  "It may be read only. ")
    main_parser.add_argument('folder_path_out', type=is_dir_path,
                             help="Path to destination folder. "
                                  "Each media file will be converted according to the instructions. "
                                  "The file hierarchy will be preserved exactly as in the original directory. ")
    main_parser.add_argument('progress_file_path',
                             # type=argparse.FileType("w+", encoding="utf-8"),
                             type=str,
                             help='Path to file with progress. It is json. ')
    main_parser.add_argument("--v_codec", type=str,  # avc=H.264, hevc=H.265
                             choices=["libx264", "h264_nvenc", "h264_amf", "h264_vaapi",
                                      "libx265", "hevc_nvenc", "hevc_amf", "h265_vaapi",
                                      "libvpx-vp9", "vp9_qsv",
                                      "libaom-av1", "av1_vaapi",
                                      "libxvid", "rawvideo"],
                             default=None, required=False,
                             help="Chosen video codec for encoding video for destination folder {folder_path_out}. ")
    main_parser.add_argument("--resolution", type=str,
                             choices=["140p", "240p", "360p", "480p", "720p", "1080p", "1440p", "2160p", "4320p",
                                      "8640p"],
                             default=None, required=False,
                             help="Chosen video resolution for destination folder {folder_path_out}. ")
    main_parser.add_argument("--fps", type=int,
                             default=None, required=False,
                             help="Chosen video FPS for destination folder {folder_path_out}. ")
    main_parser.add_argument("--interpolation", type=str,
                             choices=["dup", "accurate", "blend", "bilat", "motion", ""],
                             default=None, required=False,
                             help="Apply interpolation if you need to add new frames when increasing/decreasing FPS "
                                  "for destination folder {folder_path_out}. "
                                  "It is recommended to play with the parameter and see the results before using it.")
    main_parser.add_argument("--bitrate", type=str,
                             default=None, required=False,
                             help="Chosen video bitrate for destination folder {folder_path_out}. "
                                  "For example: 2M, 512K... ")
    main_parser.add_argument("--maxbitrate", default=False, action='store_true',
                             help="If set, the bitrate cannot exceed the value specified in flag \"--bitrate\". ")
    main_parser.add_argument("--force_bitrate", default=False, action='store_true',
                             help="If set and if the bitrate in the source file is less than in flag \"--bitrate\", "
                                  "it will still be forced to the bitrate specified in flag \"--bitrate\". "
                                  "This action will increase the size of the output file. ")

    main_parser.add_argument('--replace', action='store_true',
                             help="Replace the video if it already exists in the destination folder. ")
    main_parser.add_argument('--no-replace', dest='replace', action='store_false',
                             help="Ask if needed to replace a file in the destination folder. ")
    main_parser.set_defaults(replace=True)

    main_parser.add_argument("--stop_if_error", default=False, action='store_true',
                             help="If set and error occurred while ffmpeg is running, konvertek will exit. ")
    main_parser.add_argument("--print_command", default=False, action='store_true',
                             help="Output ffmpeg command. ")

    # problem_info
    problem_info_parser = subparsers.add_parser('problem_info', help='Print problems')
    problem_info_parser.add_argument('progress_file_path',
                                     type=str,
                                     help='Path to file with progress. It is json. ')
    # list_extensions
    list_extensions_parser = subparsers.add_parser('list_extensions', help='Print all file extensions')
    list_extensions_parser.add_argument('folder_path', type=is_dir_path, help='Path to folder. ')

    args = parser.parse_args()
    if args.command == "convert":
        if args.interpolation and not args.fps:
            parser.error("Flag --interpolation cannot be define without --fps. ")
            exit()

        if args.maxbitrate and not args.bitrate:
            parser.error("Flag --maxbitrate cannot be define without --bitrate. ")
            exit()

        if args.force_bitrate and not args.bitrate:
            parser.error("Flag --force_bitrate cannot be define without --bitrate. ")
            exit()

    return args


def is_dir_path(_string):
    if os.path.isdir(_string):
        return _string
    else:
        print(f"\"{_string}\" is not directory. ")
        raise NotADirectoryError(_string)
